/*
hijack_winhex_backup.cpp
Takes a winhex (*.whx) backup of your USB disk (assuming) and puts 
new data into it, re-using the image.
Notes:
      The original image should most likely be from the exact disk you are using.  
      Just do a winhex backup of the blank disk if you want.
6:01 PM April 23rd 2005
By Andrew Becker, AKA "Kybo Ren"
Copyright (c) 2005 Andrew Becker
*/

#include <fstream>
#include <iostream>

const unsigned int winhex_header_size = 0x17B;//for WinHex 10.75, dunno if it changed


void Error(const char *err)
{
     std::cerr << err << std::endl;
     std::cin.get();
     
     exit(1);
}

const unsigned int FileSize(std::fstream &ifile);
const unsigned int FileSize(const char *file);//overloaded ver

int main()
{
    std::string whx_filename;
    std::string bin_filename;
    std::ofstream outfile;
    std::ifstream infile;
    
    std::string buf;
    char buffer;
    
    unsigned int sz = 0;
    
    std::cout << "What is the original WHX backup of the disk called? " << std::flush;
    std::getline(std::cin, whx_filename);
    
    if(!std::cin)
    {
                 Error("Bad input.");
    }
    
    std::cout << "What is the new raw binary image name? " << std::flush;
    std::getline(std::cin, bin_filename);
    
    if(!std::cin)
    {
                 Error("Bad input.");
    }
    
    if(FileSize(bin_filename.c_str()) + winhex_header_size != FileSize(whx_filename.c_str()))
    {
                 Error("File sizes don't match.");
    }
    
    buf = "New version of " + whx_filename;
    outfile.open(buf.c_str(), std::ios::binary);
    if(!outfile)
    {
                Error("Couldn't open the new file.");
    } 
    
    infile.open(whx_filename.c_str(), std::ios::binary);
    if(!infile)
    {
                Error("Couldn't open the WHX backup file.");
    }
    
    for(unsigned int i = 0; i < winhex_header_size; ++i)
    {
                 infile.read(&buffer, 1);
                 outfile.write(&buffer, 1);
    }//write the header
    
    infile.close();
    infile.open(bin_filename.c_str(), std::ios::binary);//open the new bin file
    if(!infile)
    {
                Error("Couldn't open the new binary file.");
    }
    
    sz = FileSize(bin_filename.c_str());
    for(unsigned int i = 0; i < sz; ++i)
    {
                 infile.read(&buffer, 1);
                 outfile.write(&buffer, 1);
    }//copy the new bin file    
    
    std::cout << "Finished." << std::endl;
    std::cin.get();    
}

const unsigned int FileSize(std::fstream &ifile)
{
    unsigned int size = 0;
    unsigned int curpos = ifile.tellg();
    
    ifile.seekg(0, std::ios::end);
    size = ifile.tellg();
    
    ifile.seekg(curpos, std::ios::beg);//remember to go back to the original spot
    
    return size;
}

const unsigned int FileSize(const char *file)
{
      std::fstream temp;
      
      temp.open(file, std::ios::in | std::ios::out | std::ios::binary);
      
      if(!temp)
      {
               Error("Couldn't open temp file for FileSize().");
      }
      
      return FileSize(temp);//use our old work again
      
}
