/*
xbh2gametypes_insert.cpp
Inserts Halo 2 gametype individual files into a raw binary disk image of a memcard
Notes:
      -Files must already exist; only modifies
      -Only supports Halo 2 v1.0 saves
12:12 AM April 23rd, 2005
By Andrew Becker, AKA "Kybo Ren"
Copyright (c) 2005 Andrew Becker
*/
#include <fstream>
#include <iostream>
#include <string>
#include <limits>

const unsigned int gametype_block_length = 0x4;
const unsigned int bytes_per_block = 0x4000;//16KB per Xbox "block" :)

const unsigned int meta_section_block_num = 0x1;//start at *2nd* block
const unsigned int meta_section_length = 0x0;//UNUSED!!!!  JUST HERE FOR VANITY! :)

const unsigned int gametype_section_block_num = 0x2;//start at *3rd* block
const unsigned int gametype_section_length = 0x144;

const unsigned int image_section_block_num = 0x3;//start at 3rd block
const unsigned int image_section_length = 0x1000;//4KB in length; 1/4 of block

const unsigned int names_section_block_num = 0x0;//start at 4th block
const unsigned int names_padding_bytes_before = 0x2;//2 bytes of padding before names
const unsigned int names_padding_bytes_between = 0x40;//0x40 bytes of stuff between names

void ClearInputBuf()
{
     std::cin.ignore(std::numeric_limits<std::streamsize>::max(), '\n' );
}

void Error(const char *err)
{
     std::cerr << err << std::endl;
     std::cin.get();
     
     exit(1);
}

void GoToName(std::fstream &ifile, std::string &name);
std::string MakeFileName(std::string &folder, std::string &fname);
const unsigned int FileSize(std::fstream &ifile);
const unsigned int FileSize(const char *file);//overloaded ver

int main()
{
    unsigned int cursz = 0;
    
    std::string ofilename;
    std::string gametypename;
    std::string nametest;
    
    std::string infiles[3];
    
    std::ifstream ifile;
    std::fstream ofile;
    std::ofstream temp;
    
    unsigned int i = 0;
    
    unsigned int gametype_start_loc = 0;
    
    unsigned int image_block_size = 0;
    
    char buffer[bytes_per_block];//for copying blocks
    
    char *gametype_data = NULL;
     
    unsigned int bufptr = 0;
    unsigned int j = 0; 
    
    //Get info
    
    std::cout << "Raw binary image name: " << std::endl;
    std::getline(std::cin, ofilename);
    
    if(!std::cin)
              Error("Bad input.");
    
    std::cout << "H2 Gametype name: " << std::endl;
    std::getline(std::cin, gametypename);
    
    ofile.open(ofilename.c_str(), std::ios::in | std::ios::out | std::ios::binary);
    if(!ofile)
              Error("Couldn't open input file.");
              
    image_block_size = FileSize(ofile) / bytes_per_block;//auto-truncate :)
    if(FileSize(ofile) % bytes_per_block != 0)
              std::cout << "Warning: image file not an even number of blocks. (" << image_block_size  << ")" << std::endl;
              
    try{
        gametype_data = new char[gametype_block_length * bytes_per_block];
        }
    catch(std::bad_alloc)
    {
        Error("Couldn't allocate enough memory.");
    }//allocate memory for gametype
    
   //search for the gametype name 
    do
    {
        nametest = "";
        GoToName(ofile, nametest);
        ofile.seekg(1, std::ios::cur);
    } while(nametest != gametypename);
    ofile.seekg(-1, std::ios::cur);//we will have one left over from the do while loop
    ofile.seekg((-1 * static_cast<int>(bytes_per_block)), std::ios::cur);//go to beginning of gametype
    
    gametype_start_loc = ofile.tellg();
    ofile.seekp(ofile.tellg());//put write ptr at read ptr; i.e. the gametype loc
    
    //now we have the gametype start location we want
    
    for(i = 0; i < 4; ++i)
    {
        ofile.read(buffer, bytes_per_block);
        memcpy((gametype_data + (i*bytes_per_block)), buffer, bytes_per_block);//move the data
    }//now we have the raw gametype data in memory
    ofile.seekg(-4*bytes_per_block, std::ios::cur);
    
    bufptr = ((bytes_per_block * names_section_block_num) + names_padding_bytes_before);//point to first name; taken out of for loop because it's LONG

    for(i = 0; i < 3; bufptr += names_padding_bytes_between, ++i)
    {
        //add the names
        for(j = bufptr; static_cast<unsigned char>(gametype_data[j]) != 0xFF; ++j)
        {
                //while going through the name
                infiles[i] += gametype_data[j];//add the file name char
        }
    }
    
    //now we have the input file names

    
    //OPEN THE META FILE
    ifile.open(MakeFileName(gametypename, infiles[0]).c_str(), std::ios::in | std::ios::binary);//open the meta file for writing
    if(!ifile)
    {
              std::cout << MakeFileName(gametypename, infiles[0]).c_str() << std::endl;
              Error("Couldn't open the first file of the gametype; make sure the directory and file both exist.");
    }
    j = FileSize(MakeFileName(gametypename, infiles[0]).c_str());//reuse J as a holder of the size of the file
    bufptr = (meta_section_block_num * bytes_per_block);//should be 0, here just for vanity and maybe ease of change if the format changes
    ifile.read(buffer, j);//read the file into the buffer
    ifile.close();
    memcpy(&gametype_data[bufptr], buffer, j);//replace the file in memory
    
    //OPEN THE GAMETYPE FILE
    ifile.open(MakeFileName(gametypename, infiles[1]).c_str(), std::ios::binary);//open the meta file for writing
    if(!ifile)
    {
              Error("Couldn't open the second file of the gametype; make sure the directory and file both exist.");
    }
    j = FileSize(MakeFileName(gametypename, infiles[1]).c_str());//reuse J as a holder of the size of the file
    bufptr = (gametype_section_block_num * bytes_per_block);//the loc of the gametype
    ifile.read(buffer, j);//read the file into the buffer
    ifile.close();
    memcpy(&gametype_data[bufptr], buffer, j);//replace the file in memory
    
     //WRITE THE IMAGE FILE
    ifile.open(MakeFileName(gametypename, infiles[2]).c_str(), std::ios::binary);//open the meta file for writing
    if(!ifile)
    {
              Error("Couldn't open the third file of the gametype; make sure the directory and file both exist.");
    }
    j = FileSize(MakeFileName(gametypename, infiles[2]).c_str());//reuse J as a holder of the size of the file
    bufptr = (image_section_block_num * bytes_per_block);//the loc of the image
    ifile.read(buffer, j);//read the file into the buffer
    ifile.close();
    memcpy(&gametype_data[bufptr], buffer, j);//replace the file in memory
    
    nametest = "New version of " + ofilename;//new name of the copy
    temp.open(nametest.c_str(), std::ios::binary);
    ofile.seekg(0, std::ios::beg);   
    for(j = 0; j < gametype_start_loc; ++j)
    {
          ofile.read(buffer, 1);
          temp.write(buffer, 1);//copy a byte  
    }//copy all the data up to the gametype save location
    
    ofile.seekp((gametype_block_length * bytes_per_block), std::ios::cur);//seek to after the gametype
    
    temp.write(gametype_data, (gametype_block_length * bytes_per_block));//write the NEW data
    
    cursz = FileSize(ofilename.c_str());     
    for(j = ofile.tellg(); j <= cursz; ++j )
    {
          ofile.read(buffer, 1);
          temp.write(buffer, 1);    
    }//copy all the data past the gametype save location
    
    delete [] gametype_data;//deallocate
    
    std::cout << "Finished." << std::endl;
    std::cin.get();
}

std::string MakeFileName(std::string &folder, std::string &fname)
{
    return folder + '/' + fname;
}

const unsigned int FileSize(std::fstream &ifile)
{
    unsigned int size = 0;
    unsigned int curpos = ifile.tellg();
    
    ifile.seekg(0, std::ios::end);
    size = ifile.tellg();
    
    ifile.seekg(curpos, std::ios::beg);//remember to go back to the original spot
    
    return size;
}

const unsigned int FileSize(const char *file)
{
      std::fstream temp;
      
      temp.open(file, std::ios::in | std::ios::out | std::ios::binary);
      
      if(!temp)
      {
               Error("Couldn't open temp file for FileSize().");
      }
      
      return FileSize(temp);//use our old work again
      
}


void GoToName(std::fstream &ifile, std::string &name)
{
    bool found = false;
    char block[64];
    unsigned int i = 0;
    
    while(!found)
    {
        if(ifile.get() == 0xFF &&
        ifile.get() == 0xFE)
        {
                ifile.read(block, 64);
                if(
                block[0] == 'N' &&
                block[2] == 'a' &&
                block[4] == 'm' &&
                block[6] == 'e' &&
                block[8] == '=')
                {      
                                for(i = 10; block[i] != ' '; i+=2);
                                i+=2;//go to the gametype name
                   
                                for(; block[i] != 0x0D && i < 64; i+=2)
                                {
                                     name += block[i];
                                }//give name
                   
                                found = true;
                                ifile.seekg(-66, std::ios::cur);
                                return;
                }
        }
        
        if(!ifile)
        {
        return;
        }
    }       
}

const unsigned int MetaSectionLength(char *gamedata)
{
      unsigned int i;
      
      for(i = 1; i < (gametype_block_length * bytes_per_block); ++i)//start at *1* because first byte is 0xFF ;)
      {
                   if(static_cast<unsigned char>(gamedata[i]) == 0xFF)
                   {
                                  break;
                   }
      }
      
      return i;
}
